#!/usr/bin/env python 
"""This submodule defines a collection of container classes that are used through the library
"""
# load the modules: 
from __future__ import annotations
from Bio import SeqIO 
import os
import pandas as pd 
from typing import List, Dict 
from Bio import SeqIO
# define the class 
class SeqDB: 
	"""Load a FASTA file and constructs a lock up dictionary where sequence ids are  keys and sequences are values. 
	"""
	def __init__(self, path2fasta: str) -> SeqDB: 
		"""The class constructor 

		:param path2fasta: the path to load the fasta database 
		:type path2fasta: str
		:raises FileNotFoundError: In case the provided path does not exist or the file can not be found. 
		:return: a sequence database instance
		:rtype: SeqDB
		"""
		if not os.path.exists(path2fasta): 
			raise FileNotFoundError(f"Your provided path: {path2fasta} does not exist!")
		# define a private variable to hold the results 
		self._seqs=dict() 
		seq_gen=SeqIO.parse(path2fasta,'fasta')
		for seq in seq_gen: 
			key=seq.id.split('|')[1]
			self._seqs[key]=str(seq.seq) 
		return

	def get_seq(self,protein_id:str)->str: 
		"""returns the corresponding sequence if the provided protein-id is defined in the database.

		:param protein_id: The protein id to retrive its sequence, CASE SENSITIVE!!. 
		:type protein_id: str
		:raises KeyError: If the provided protein does not exist in the database 
		:return: the protein sequence
		:rtype: str
		"""
		if protein_id not in self._seqs.keys(): 
			raise KeyError(f"Your provided protein id:{protein_id} is not in the database" )
		return self._seqs[protein_id] 

	def __len__(self)->int:	
		"""A magic function to return the number of the sequence in the database 

		:return: the number of the sequences in the database 
		:rtype: int
		"""
		return len(self._seqs)

	def __getitem__(self,protein_id: str)->str:
		"""a magic function for indexing through the elements of the library.   

		:param protein_id: the protein id 
		:type protein_id: str
		:return: the sequence of the provided id 
		:rtype: str
		"""
		return self.get_seq(protein_id)
	
	def __str__(self)->str:
		"""return a string representation for the class 

		:return: return a string representation for the class
		:rtype: str
		"""
		return f'A sequence database with {len(self)} sequence'

	def __repr__(self)->str:
		"""return the representation of the database instance.

		:return: a string representation of the class 
		:rtype: str
		"""
		return str(self)
	
	def has_sequence(self, sequence_id: str)->bool:
		"""check if the provided sequence id is an element of the database or not		

		:param sequence_name: The id of the sequence, CASE SENSITIVE!!. 
		:type sequence_name: str
		:return: True if the database has this id, False otherwise. 
		:rtype: bool
		"""
		return sequence_id in self._seqs.keys()

class CellularLocationDB:
	"""The class provides an API to access the cellular location information from a database \
	that follows the structure of the Human Proteome Atlas sub-cellular location database. See https://www.proteinatlas.org/about/download \
	for more details. 
	"""
	def __init__(self, 
				path2data: str = "https://www.proteinatlas.org/download/subcellular_location.tsv.zip",
				sep: str = '\t')->CellularLocationDB:
		"""	The class constructor

		:param path2data: the path to load the database or the URL to download it. defaults to https://www.proteinatlas.org/download/subcellular_location.tsv.zip
		:type path2data: str
		:param sep: The table seperator, defaults to '\t'
		:type sep: str, optional
		:raises IOError: Incase the provided table could not be loaded 
		:return: an instance of the class 
		:rtype: CellularLocationDB
		"""
		try: 
			self._table: pd.DataFrame = pd.read_csv(path2data,sep=sep)
		except Exception as exp: 
			raise IOError(f'While loading the sub cellular location dataset the following error: {exp} was encountered.')
	
	def get_genes(self)-> List[str]:
		"""return a list of all gene ids in the dataset 

		:return: all genes ids currently defined in the database 
		:rtype: List[str]
		"""
		return self._table.iloc[:,0].tolist()
	
	def get_gene_names(self)->List[str]:
		"""return a list of all gene names in the dataset 

		:return: the names of all genes in the database 
		:rtype: List[str]
		"""
		return self._table.iloc[:,1].tolist()
	
	def get_main_location(self, gene_id: str = None, corresponds= None)->List[str]:
		"""Return the main location(s) of the provided gene id or gene name. \ 
		If both gene Id and gene name are provided, gene_id has a higher precedence 

		:param gene_id: The id of the gene of interest, defaults to None
		:type gene_id: str, optional
		:param gene_name: The name of the gene of interest, defaults to None
		:type gene_name: [type], optional
		:raises ValueError: if both gene_id and gene_name are None
		:raises KeyError: if gene_id is None and gene_name is not in the database 
		:raises KeyError: if gene_name is None and gene_id is not in the database 
		:raises RuntimeError: Incase an error was encountered while retriving the element from the database
		:return: the main location where the protein that corresponds to the provided name or id is located. 
		:rtype: List[str]
		"""
		if gene_id is None and corresponds is None:
			raise ValueError(f'gene_name and gene_id can not be None')
		
		if gene_id is not None:
			if gene_id not in self.get_genes():
				raise KeyError(f'The provided gene id: {gene_id} is not in the database')
		
		if corresponds is not None:
			if corresponds not in self.get_gene_names():
				raise KeyError(f'The provided gene name : {corresponds} is not in the database')
		# get the main location 
		try:
			if gene_id is not None:  
				main_location: str = self._table.loc[self._table.iloc[:,0]==gene_id,]['Main location'] 
			else: 
				main_location: str = self._table.loc[self._table.iloc[:,1]==corresponds,]['Main location'] 
		except Exception as exp: 
			raise RuntimeError(f'While getting the main cellular location the following error was encounter: {exp}')
		# split the locations 
		if main_location.isna().any():
			locations: List[str] = ['UNK']
		else: 
			locations: List[str] = main_location.tolist()[0].split(';')
		# return the resuts 
		return locations 
	
	def get_approved_location(self, gene_id: str = None, gene_name= None) -> List[str]:
		"""return the location of the provided gene id or gene name 

		:param gene_id: the id of the gene of interest, defaults to None
		:type gene_id: str, optional
		:param gene_name: the name of gene of interest, defaults to None
		:type gene_name: [type], optional
		:raises ValueError: if both gene_id and gene_name are None
		:raises KeyError: if gene_id is None and gene_name is not in the database 
		:raises KeyError: if gene_name is None and gene_id is not in the database 
		:raises RuntimeError: Incase an error was encountered while retriving the element from the database. 
		:return: The approved location where the protein that corresponds to the provided name or id is located. 
		:rtype: List[str]
		"""
		if gene_id is None and gene_name is None:
			raise ValueError(f'gene_name and gene_id can not be None')
		
		if gene_id is not None:
			if gene_id not in self.get_genes():
				raise KeyError(f'The provided gene id: {gene_id} is not in the database')
		
		if gene_name is not None:
			if gene_name not in self.get_gene_names():
				raise KeyError(f'The provided gene name : {gene_name} is not in the database')
		
		# get the approved locations 
		try:
			if gene_id is not None:  
				approved_locations: str = self._table.loc[self._table.iloc[:,0]==gene_id,]['Approved']  
			else: 
				approved_locations: str = self._table.loc[self._table.iloc[:,1]==gene_name,]['Approved']
		except Exception as exp: 
			raise RuntimeError(f'While getting the approved locations the following error was encounter: {exp}')
		# split the locations 
		if approved_locations.isna().any(): 
			locations: List[str] =  ['UNK']
		else: 
			locations: List[str] =  approved_locations.tolist()[0].split(';')	
		# return the resuts 
		return locations 
	
	def get_go_names(self,gene_id: str = None, gene_name= None )->List[str]:
		"""return the location of the provided gene id or gene name 

		:param gene_id: the id of the gene of interest , defaults to None
		:type gene_id: str, optional
		:param gene_name: the name of the gene of interest , defaults to None
		:type gene_name: [type], optional
		:raises ValueError: if both gene_id and gene_name are None
		:raises KeyError: if gene_id is None and gene_name is not in the database 
		:raises KeyError: if gene_name is None and gene_id is not in the database 
		:raises RuntimeError: incase an error was encountered while retriving the element from the database.
		:return: The gene ontology, GO, location where the protein that corresponds to the provided name or id is located. 
		:rtype: List[str]
		"""
		if gene_id is None and gene_name is None:
			raise ValueError(f'gene_name and gene_id can not be None')
		
		if gene_id is not None:
			if gene_id not in self.get_genes():
				raise KeyError(f'The provided gene id: {gene_id} is not in the database')
		
		if gene_name is not None:
			if gene_name not in self.get_gene_names():
				raise KeyError(f'The provided gene name : {gene_name} is not in the database')
		
		# get gene ontology IDs
		try:
			if gene_id is not None:  
				go_ids : str = self._table.loc[self._table.iloc[:,0]==gene_id,]['GO id'] 
			else: 
				go_ids: str = self._table.loc[self._table.iloc[:,1]==gene_name,]['GO id']
		except Exception as exp: 
			raise RuntimeError(f'While getting the approved locations the following error was encounter: {exp}')
		# split the locations 
		if go_ids.isna().any(): 
			_ids: List[str] =  ['UNK']
		else: 
			_ids: List[str] =  go_ids.tolist()[0].split(';')	
		# return the resuts 
		return _ids

	def get_table(self) -> pd.DataFrame:
		"""return the instance table 

		:return: the location table of the instance. 
		:rtype: pd.DataFrame
		"""
		return self._table

	def add_to_database(self,genes_to_add: CellularLocationDB )->None:
		"""adds the the location of more proteins to the database. 

		:param genes_to_add: a CellularLocationDB instance containing the genes that shall be added to the database.   
		:type genes_to_add: CellularLocationDB
		:raises ValueError: if the genes_to_add to the database are already defined in the database
		:raises RuntimeError: incase any other error has been encountered while merging the tables.
		"""
		# check that the genes in provided database do not exist in the current instance.
		genes: List[str] = genes_to_add.get_genes()

		for gene in genes: 
			if gene in self.get_genes():
				raise ValueError(f'The provided gene: {gene} is already defined in current instance table')
		# add the genes to the database, concat the databases  
		try: 
			self._table=pd.concat([self._table, genes_to_add.get_table()],axis=1)
		except Exception as exp: 
			raise RuntimeError(f'While combing the database, the following error was encountered: {exp}')
	
	# define some magic functions 
	def __len__(self)->int:
		"""return the number of unique genes in the current instance
		
		:return: the number of unique genes in the current instance
		:rtype: int
		"""
		return len(set(self._table.iloc[:,0]))
		
	def __str__(self)->str: 
		""" return a string representation of the instance 

		:return: a string representation for the instance 
		:rtype: str
		"""
		return f' A sub-cellular localization database covering {len(self)} genes'

	def __repr__(self)->str: 
		return str(self)


class GeneExpressionDB: 
	"""The class provides an API to access gene expression data stored in table that follows the same structure as \ 
	the Human proteome Atlas Normalized RNA Expression see  https://www.proteinatlas.org/about/download for more details 
	"""
	def __init__(self, path2data: str = 'https://www.proteinatlas.org/download/rna_tissue_consensus.tsv.zip', 
					sep: str = '\t')->GeneExpressionDB: 
		"""The class constructor 

		:param path2data: The path to the expression table, defaults to, 'https://www.proteinatlas.org/download/rna_tissue_consensus.tsv.zip'. 
		:type path2data: str, optional 
		:param sep: The separator for the input table, defaults to '\t'
		:type sep: str, optional
		:raises IOError: Incase the provided table could not be load. 
		:return: an instance of the class 
		:rtype: GeneExpressionDB
		"""
		try: 
			self._table: pd.DataFrame = pd.read_csv(path2data,sep=sep)
		except Exception as exp: 
			raise IOError(f'While loading the RNA expression dataset the following error: {exp} was encountered.')

	def get_genes(self)->List[str]:
		"""returns a list of the UNIQUE gene ids currently in the database. 

		:return: The list of the UNIQUE gene ids currently in the database 
		:rtype: List[str]
		"""
		return list(set(self._table.iloc[:,0].tolist()))
	
	def get_gene_names(self)->List[str]:
		"""returns a list of the UNIQUE gene names currently in the database 

		:return: A list of the UNIQUE gene names currently in the database 
		:rtype: List[str]
		"""
		return list(set(self._table.iloc[:,1].tolist()))
	
	def get_tissues(self)->List[str]:
		"""return a list of the tissues in the current database
		
		:return: A list containing the names of the UNIQUE tissues in the database.  
		:rtype: List[str]
		"""
		return list(set(self._table.Tissue))

	def get_expression(self, gene_name: str = None, gene_id: str = None) ->pd.DataFrame: 
		"""Return a table summarizing the expression of the provided gene name or gene id accross different tissues. 

		:param gene_id: the id of the gene of interest, defaults to None
		:type gene_id: str, optional
		:param gene_name: the name of the gene of interest, defaults to None
		:type gene_name: [type], optional
		:raises ValueError: if both gene_id and gene_name are None
		:raises KeyError: if gene_id is None and gene_name is not in the database 
		:raises KeyError: if gene_name is None and gene_id is not in the database 
		:raises RuntimeError: incase an error was encountered while retriving the elements from the database
		:return: A table summarizing the expression of the provided gene accross all tissues in the database
		:rtype: pd.DataFrame
		"""
		if gene_id is None and gene_name is None:
			raise ValueError(f'gene_name and gene_id can not be None')
		
		if gene_id is not None:
			if gene_id not in self.get_genes():
				raise KeyError(f'The provided gene id: {gene_id} is not in the database')
		
		if gene_name is not None:
			if gene_name not in self.get_gene_names():
				raise KeyError(f'The provided gene name : {gene_name} is not in the database')
		
		if self._table.loc[self._table.iloc[:,0]==gene_name]['Gene name']!=gene_name:
			raise ValueError('The provided gene names: {gene_name} does not match the provided gene id: {gene_id}')
		# get the expression accross all tissues 			
		try: 
			if gene_id is not None:  
				return self._table.loc[self._table.iloc[:,0]==gene_id]
			else:
				return self._table.loc[self._table.iloc[:,1]==gene_name]
		except Exception as exp:
			raise RuntimeError(f'The following error has been encountered while prepearing the table: {exp}') 
			
	def get_expression_in_tissue(self, tissue_name:str)-> pd.DataFrame:
		"""return the expression profile of the provided tissue  

		:param tissue_name: The name of the tissue 
		:type tissue_name: str
		:raises KeyError: Incase the provided tissue is not defined in the database
		:raises RuntimeError: In case an error was encountered while generating the expression profile. 
		:return: A table summarizing the expression of all genes in the provided tissue.
		:rtype: pd.DataFrame
		"""
		if tissue_name not in self.get_tissues():
			raise KeyError(f"The provided tissue: {tissue_name} is not support, currently supported tissue are:  {','.join(self.get_tissues())} ")
		# return the table 
		try: 
			table: pd.DataFrame = self._table.loc[self._table.iloc[:,2]==tissue_name]
			res: pd.DataFrame = table[['Gene','Gene name','NX']]
			return res
		except Exception as exp:
			raise RuntimeError(f'While loading the database the following error was encountered: {exp}')
	
	def get_table(self)->pd.DataFrame:
		"""return a table containing the expression value of all the genes accross all tissues in the current instance

		:return: The expression of all genes accross all tissues in the database. 
		:rtype: pd.DataFrame
		"""
		return self._table[['Gene', 'Gene name', 'NX']]
		
	def __len__(self)->int: 
		"""return the number of tissues in the database

		:return: the number of tissues in the database
		:rtype: int
		"""
		return len(self.get_tissues())
	
	def __str__(self)-> str:
		"""A string representation of the class

		:rtype: str
		"""
		return f'an expression database covering: {len(self.get_genes())} genes accross {len(self)} tissues'
	
	def __repr__(self) ->str: 
		return str(self)

class OrganismDB:
	"""Extract information about the source organsim of a collection of protein sequences\ 
	from a fasta file and provides an API to query the results. \
	The function expect the input fasta file to have headers written in the UNIPROT format. 
	"""
	def __init__(self,path2Fasta:str)->OrganismDB:
		""" The class constructor 
	
		:param path2Fasta: The path to a fasta sequence database to obtain the protein sequences
		:type path2Fasta: str
		:raises IOError: Incase the database could not be loaded 
		:return: A new OrganismDB instance 
		:rtype: OrganismDB
		"""
		try: 
			seq_gene=SeqIO.parse(path2Fasta,'fasta')
		except Exception as exp: 
			raise IOError(f'While loading your input database: {path2Fasta}, the following error was encountered: {exp}')
		# define a dict that hold the map 
		self._map: Dict[str,str] = dict()
		# fill the elements in the map 
		for seq in seq_gene: 
			temp_name_org: List[str]=seq.id.split('|')
			self._map[temp_name_org[1]]= temp_name_org[2].split('_')[1]
	
	def get_unique_orgs(self)->List[str]:
		"""Get the number of unique organisms in the database
		
		:return: a list of all unique organisms in the current instance 
		:rtype: List[str]
		"""
		return list(set(self._map.values()))
	
	def get_number_protein_per_organism(self)->pd.DataFrame:
		"""Provides a table containing the number of proteins per organism. 
		
		:return: A table containing the number of proteins per organism
		:rtype: pd.DataFrame
		"""
		unique_orgs: List[str] = self.get_unique_orgs()
		# create a dict counter 
		counter: Dict[str,int]= dict()
		# Initialize the counts to 0s 
		for org in unique_orgs: 
			counter[org]=0
		# fill the dict with counts 
		for key in self._map.keys(): 
			counter[self._map[key]]+=1
		# chaning the map from int -> List[int] to fit the dataframe requirements 
		for key in counter: 
			counter[key]=[counter[key]]
		# create the dataframe 
		res: pd.DataFrame = pd.DataFrame(counter).T 
		# add the org-names as keys  
		res['Names'] = res.index.tolist()
		# return the results 
		return res 
	
	def get_org(self,prot_id:str)->str: 
		"""return the parent organism of the provided protein identifer

		:param prot_id: the id of the protein of interest 
		:type prot_id: str
		:raises KeyError: incase the provided identifier is not in the database 
		:return: the name of the parent organism, i.e. the source organism. 
		:rtype: str
		"""
		if prot_id not in self._map.keys(): 
			raise KeyError(f'The provided key is not defined in the database')
		return self._map[prot_id]
	
	def __len__(self)->int: 
		"""return the number of protein id parent organism pairs currently stored  in the database.
		
		:rtype: int
		"""
		return len(self._map)
	
	def __str__(self)->str: 
		""" return a string representation of the class 
				:rtype: str
		"""
		return str(f'An organism database with: {len(self)} entry.')

	def __repr__(self)->str: 
		return str(self)

		

	



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="IPTKL", 
    version="0.4.2",
    author="Hesham ElAbd",
    author_email="h.elabd@ikmb.uni-kiel.de",
    description="IPTK is a Pythonic library specialized in the analysis of HLA-peptidomes identified through an Immunopeptiomics pipeline.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ikmb/iptoolkit",
    packages=setuptools.find_packages(),
    install_requires=['numpy','pandas','biopython','seaborn' ,'matplotlib', 'plotly' ,'mhcnames', 
                'pyteomics', 'h5py', 'logomaker', 'colour', 'lxml', 'nglview', 'sklearn', 'scipy',
                'statannot'], 
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)



#!/usr/bin/env python 
"""An ExperimentalSet which is a collection of experiments.
The class provides an API for integrating and comparing different experimental instances.   
"""
# load the modules 
from __future__ import annotations
import numpy as np 
import pandas as pd
from IPTK.Classes.Experiment import Experiment 
from IPTK.Classes.Peptide import Peptide
from IPTK.Analysis.AnalysisFunction import get_binnary_peptide_overlap, get_binnary_protein_overlap 
from IPTK.Analysis.AnalysisFunction import compute_change_in_protein_representation
from IPTK.Analysis.AnalysisFunction import compute_expression_correlation
from typing import Dict, List 
## define some types 
Experiments=Dict[str,Experiment]
Names=List[str]
Counts=Dict[str,int ]
Peptides=List[Peptide]
Proteins=List[str]
# define the class 
class ExperimentSet: 
    """an API for integrating and comparing different experimental instances
    """
    def __init__(self,**exp_id_pair)->ExperimentSet:
        """Create a new instance using an arbitrary number of experiments. 
        """
        # an experimental dict 
        self._exps=dict()
        # if the number of the provided experiments is zero, we create an an empty instance that will be filled later 
        if len(exp_id_pair)==0: return 
        # get the key and value from the exp-id-pair 
        for ident, exp in exp_id_pair.items():
            if isinstance(exp,Experiment): 
                self._exps[ident]=exp
            else:
                raise ValueError(f"""The constructor expected the value in your input pair to be an instance of class Experiment,
                                     however, your key: {ident} has a type of {type(exp)}""")
    
    def add_experiment(self, **exp_id_pair)->None:
        """adds an arbitrary number of experiments to the set 
        """
        for ident, exp in exp_id_pair.items():
            if isinstance(exp,Experiment): 
                self._exps[ident]=exp
            else:
                raise ValueError(f"""The constructor expected the value in your input pair to be an instance of class Experiment,
                                     however, your key: {ident} has a type of {type(exp)}""")

    def get_num_experiments_in_the_set(self)->int:
        """
        :return: The number of experiments currently in the set 
        :rtype: int
        """
        return len(self._exps)
    
    def get_experiments(self)->Dict[Experiment]:
        """
        :return: returns a dict with all the experiments stored in the instance as value and ids as keys. 
        :rtype: Dict[Experiment]
        """
        return self._exps
    
    def get_experiment(self, exp_name:str)->Experiment:
        """ returns the experiment pointed to by the provided experimental name

        :param exp_name: the name of the experiment 
        :type exp_name: str
        :raises KeyError: if the provided experimental name is not in the dataset. 
        :return: the experiment corresponds to the provided name 
        :rtype: Experiment
        """
        # define the experiment name 
        if exp_name not in self._exps.keys():
            raise KeyError(f"The provided experimental name: {exp_name} is not defined in the current experiment") 
        # return the experiment 
        return self._exps[exp_name]

    def get_experimental_names(self)->Names:
        """
        :return: A list with all the identifiers of the experiments in the set  
        :rtype: Names
        """
        return list(self._exps.keys())
    
    def get_unique_orgs(self)->List[str]:
        """
        :return: A list of the unique organisms in the set
        :rtype: List[str]
        """
        unique_orgs: List[str] = []
        for name in self.get_experimental_names():
             unique_orgs.extend(self._exps[name].get_orgs())
        # compute the unique organisms in the list of organisms 
        unique_orgs=list(set(unique_orgs))
        return unique_orgs

    def get_total_peptide_per_org_count(self) ->pd.DataFrame:
        """
        :return: The total count of peptides per organism accross the all experiments in the set. 
        :rtype: pd.DataFrame
        """
        # first, get the unique organisms in the set.  
        unique_orgs: List[str] = self.get_unique_orgs()
        # create a counter and initialize it to zero to hold the results 
        org_counter=dict()
        for org in unique_orgs:
            org_counter[org]=0
        # update the counts 
        for name in self.get_experimental_names():
            for _, row in self._exps[name].get_peptides_per_organism().iterrows():
                 org_counter[row['Organisms']]+=row['Counts']
       	# make the data compatible with data frames 
        for org in org_counter.keys():
            org_counter[org]=[org_counter[org]]
		# create a dataframe
        res: pd.DataFrame = pd.DataFrame(org_counter).T
		# add the index as an extra-columns 
        res['Organisms'] = res.index.tolist()
        res.columns=['Counts','Organisms']
        # reformat the dataframe 
        res.reset_index(drop=True,inplace=True)
        res=res.reindex(columns=['Organisms','Counts'])
        # sort the results 
        res=res.sort_values(by='Counts',ascending=False)
        return res

    def compare_org_count_among_exps(self, org:str, abs_count: bool =False) ->pd.DataFrame: 
        """
        :param org: The name of the organism to query the database for it. 
        :type org: str
        :param abs_count: The absolute count, defaults to False
        :type abs_count: bool, optional
        :return: The count of the peptides that belong to a specific organism in the database.
        :rtype: pd.DataFrame
        """
        # allocate and array to hold the results 
        res: np.ndarray = np.zeros((len(self),len(self)))
        # loop over all the experiments in the set 
        # initialize the counters 
        experimental_name: List[str] = self.get_experimental_names()
        for row_idx in range(len(experimental_name)):
            # get the counts per column 
            org_row: pd.DataFrame = self._exps[experimental_name[row_idx]].get_peptides_per_organism()
            org_row_count: int = org_row.loc[org_row.iloc[:,0]==org]['Counts'] 
            if org_row_count.empty:
                org_row_count=0
            else:
                org_row_count=org_row_count.tolist()[0]
            # get the counts per row 
            for col_idx in range(len(experimental_name)):
                # get the row-column
                org_col: pd.DataFrame = self._exps[experimental_name[col_idx]].get_peptides_per_organism()
                org_col_count: int = org_col.loc[org_col.iloc[:,0]==org]['Counts'] 
                if org_col_count.empty: 
                    org_col_count=0
                else: 
                    org_col_count=org_col_count.tolist()[0]
                # add the count to the
                res[row_idx,col_idx]= org_row_count-org_col_count
        # return a data frame of the results 
        res=pd.DataFrame(res)
        # add the name of columns and index 
        res.columns=experimental_name
        res.index=experimental_name
        # return the results
        return res 
    
    def drop_peptides_belong_to_org(self, org_name: str) -> None:
        """drops all the peptides that belong to the provided organisms from all experiments in the set.


        :param org_name: The name of the organism to drop.
        :type org_name: str
        """
        for name in self.get_experimental_names():
            self[name].drop_peptide_belong_to_org(org_name)
        return 

    def __getitem__(self, name:str)->Experiment:
        """ A magic function for accessing the experiments stored in the set

        :param name: The experiment name or id 
        :type name: str
        :raises KeyError: if the provided name is not defined in the current instance. 
        :return: the experiment with the corresponding name
        :rtype: Experiment
        """
        if name in self.get_experimental_names():
            return self._exps[name]
        else:
            raise KeyError(f"Name: {name} is not defined in the current instance")
    
    def __len__(self)->int:
        """  A magic function for computing the length of the set, which is the number of experiments 
        stored inside the instance 
        
        :return: The number of experiments stored inside the instance 
        :rtype: int 
        """
        return len(self._exps)
    
    def __str__(self)->str:
        """ A string representation of the class

        :return: A string representation of the class
        :rtype: str
        """
        return f'an experimental set with {len(self)} Experiments in it'
        
    
    def get_tissue_counts(self)->Counts:
        """
        :return: The number of experiments obtained from each tissue in the current instance
        :rtype: Counts
        """
        tissues_counts=dict()
        # loop over all the elements in the set 
        for exp in self._exps.keys(): 
            tissue_=self._exps[exp].get_tissue_name()
            if tissue_ in tissues_counts.keys():
                tissues_counts[tissue_]+=1
            else: 
                tissues_counts[tissue_]=1
        # return the results 
        return tissues_counts
    
    def get_allele_count(self)->Counts:
        """
        :return: The number of experiments obtained from each allele in the instance.
        :rtype: Counts
        """
        allele_counts=dict()
        # loop over all the elements in the set 
        for exp in self._exps.keys(): 
            alleles=self._exps[exp].get_hla_allele()
            for allele in alleles:
                if allele in allele_counts.keys():
                    allele_counts[allele]+=1
                else: 
                    allele_counts[allele]=1
        # return the results 
        return allele_counts
    
    def get_proband_count(self)->Counts:
        """
        :return: The number of experiments obtained from each proband in the ExperimentalSet.
        :rtype: Counts
        """
        proband_count=dict()
        # loop over all the elements in the set 
        for exp in self._exps.keys(): 
            proband_=self._exps[exp].get_proband_name()
            if proband_ in proband_count.keys():
                proband_count[proband_]+=1
            else: 
                proband_count[proband_]=1
        # return the gene_counts 
        return proband_count
    
    def group_by_tissue(self)->Dict[str,ExperimentSet]:
        """
        :return:  A map between each tissue and an ExperimentalSet object representing all experiments belonging to that tissue. 
        :rtype: Dict[str,ExperimentSet]
        """
        # define the set of tissues 
        tissues2exps: Dict[str,ExperimentSet]=dict()
        tissue_counter: Dict[str, int]=dict()
        # Initialize the counters 
        tissues: List[str] =list(set([self._exps[exp].get_tissue_name() for exp in self._exps.keys()]))
        for tissue in tissues:
            tissue_counter[tissue]=0
        # loop over all the elements in the set 
        for exp in self._exps.keys(): 
            tissue_=self._exps[exp].get_tissue_name()
            if tissue_ in tissues2exps.keys():
                temp_pair: Dict[str,Experiment]={
                    tissue_+str(tissue_counter[tissue_]):self._exps[exp]
                }
                tissues2exps[tissue_].add_experiment(**temp_pair)
                tissue_counter[tissue_]+=1
            else:
                temp_pair: Dict[str,Experiment]={
                    tissue_+str(tissue_counter[tissue_]):self._exps[exp]
                } 
                tissues2exps[tissue_]=ExperimentSet(**temp_pair)
                tissue_counter[tissue_]+=1
        # return results 
        return tissues2exps
    
    def group_by_proband(self)->Dict[str,ExperimentSet]:
        """
        :return: a map between each proband and an Experimentalset object representing all experiments objects \
        belonging to each unique proband in the set. 
        :rtype: Dict[str,ExperimentSet]
        """
        # define the set of tissues 
        proband2exps: Dict[str, ExperimentSet]=dict()
        proband_counter: Dict[str,int]=dict() 
        # define the probands 
        probands: List[str] = list(set([self._exps[exp].get_proband_name() for exp in self._exps.keys()]))
        for proband in probands: 
            proband_counter[proband]=0
        # loop over all the elements in the set 
        for exp in self._exps.keys(): 
            proband_=self._exps[exp].get_proband_name()
            if proband_ in proband2exps.keys():
                temp_pair: Dict[str,Experiment]={
                    proband_+str(proband_counter[proband_]):self._exps[exp]
                }
                proband2exps[proband_].add_experiment(**temp_pair)
                proband_counter[proband_]+=1
            else:
                temp_pair: Dict[str,Experiment]={
                    proband_+str(proband_counter[proband_]):self._exps[exp]
                }
                proband2exps[proband_]=ExperimentSet(**temp_pair)
                proband_counter[proband_]+=1
        # return results 
        return proband2exps

    
    def get_unique_peptides(self)->Peptides:
        """
        :return: The set of unique peptides in the experimentalSet 
        :rtype: Peptides
        """
        res:List[str]=[]
        for exp_name in self.get_experimental_names(): 
            res.extend(self[exp_name].get_peptides())
        return list(set(res))
    
    def get_unique_proteins(self)->Proteins:
        """
        :return: The set of unique proteins in the experimentalset
        :rtype: Proteins
        """
        res:List[str]=[]
        for exp in self.get_experimental_names():
            res.extend(self[exp].get_proteins())
        return list(set(res))
    
    def is_peptide_present_in_all(self, peptide:str)->bool:
        """
        :param peptide: The peptide sequence to search its occurrences in every experiment contained in the set
        :type peptide: str
        :return: True if peptide is present in all experiments inside the instance, False otherwise 
        :rtype: bool
        """
        # check whether the peptide is defined in the set of unique peptides or not  
        if peptide not in self.get_unique_peptides():
            return False
        # check if the peptide is a member of ALL EXPERIMENTs in the set 
        for exp_name in self.get_experimental_names(): 
            if not self[exp_name].is_member(peptide):
                return False 
        # if the peptide is a member of every experiment in the set, we return True
        return True
    
    def is_protein_present_in_all(self, protein:str)->bool:
        """
        :param protein: the protein id to search its occurrences in every experimental in the set 
        :type protein: str
        :return: True if peptide is present in all experiments inside the instance, False otherwise 
        :rtype: bool
        """
        if protein not in self.get_unique_proteins():
            return False
        #check that the protein is a member of every experiment 
        for exp_name in self.get_experimental_names():
            if not self[exp_name].is_a_parent_protein(protein):
                return False
        # return true as it is defined in all the experiment in the dataset. 
        return True

    def get_peptides_present_in_all(self)->Peptides:
        """
        :return: The peptides that are observed in every experiment in the set.
        :rtype: Peptides
        """
        all_peptides=self.get_unique_peptides()
        results=[]
        for pep in all_peptides: 
            if self.is_peptide_present_in_all(pep):
                results.append(pep)
        return results 
    
    def get_proteins_present_in_all(self)->Proteins:
        """
        :return: The proteins that are inferred in all experiments of the set. 
        :rtype: Proteins
        """
        all_proteins=self.get_unique_proteins()
        results: List[str]=[]
        for protein in all_proteins: 
            if self.is_protein_present_in_all(protein):
                results.append(protein)
        return results 
        
    def compute_peptide_overlap_matrix(self)->np.ndarray:
        """

        :return: A 2D matrix containing the number of peptide overlapping between each pair of experiments inside the current instance collection of experiment.   
        :rtype: np.ndarray
        """
        # allocate the results array 
        results_array=np.zeros(shape=(len(self), len(self))) 
        experiment_names=self.get_experimental_names()
        for raw_idx in range(len(experiment_names)):
            for col_idx in range(len(experiment_names)):
                results_array[raw_idx,col_idx]=len(
                    get_binnary_peptide_overlap(self[experiment_names[raw_idx]],
                                                self[experiment_names[col_idx]])) # compute the peptide overlap    
        # construct a dataframe from the results 
        results_df=pd.DataFrame(results_array)      
        # add the columns and index to the results df 
        results_df.columns=experiment_names
        results_df.index=experiment_names
        # return the results 
        return results_df 

    def compute_protein_overlap_matrix(self)->np.ndarray:
        """
        :return: returns a 2D matrix containing the number of proteins overlapping between each pair of experiments inside the current instance collection of experiment.  
        :rtype: np.ndarray
        """
        # allocate the results array 
        results_array=np.zeros(shape=(len(self), len(self))) 
        experiment_names=self.get_experimental_names()
        for raw_idx in range(len(experiment_names)):
            for col_idx in range(len(experiment_names)):
                results_array[raw_idx,col_idx]=len(
                    get_binnary_protein_overlap(self[experiment_names[raw_idx]],
                                                self[experiment_names[col_idx]])) # compute the peptide overlap    
        # construct a dataframe from the results 
        results_df=pd.DataFrame(results_array)      
        # add the columns and index to the results df 
        results_df.columns=experiment_names
        results_df.index=experiment_names
        # return the results 
        return results_df 
    
    def compute_peptide_representation_count(self)->Counts: 
        """
        :return: The number of times a peptide was observed accross experiments in the set 
        :rtype: Counts
        """
        # define the results object 
        results: Dict[str, int]=dict()
        unique_peptides: List[str]=self.get_unique_peptides()
        # fill the dictionary and initialize the counts to zeros 
        for peptide in unique_peptides:
            results[peptide]=0
        # loop over all the experiment to count the peptides 
        for peptide in unique_peptides:
            for exp_name in self.get_experimental_names():
                if self[exp_name].is_member(peptide):
                    results[peptide]+=1
        # return the results after filling it 
        return results 
    
    def compute_protein_representation_count(self)->Counts: 
        """
        :return: The number of times a protein was observed accross the experiment in the set
        :rtype: Counts
        """
        results=dict()
        unique_proteins=self.get_unique_proteins()
        # fill the dictionary and initialize the counts to zeros 
        for prot in unique_proteins:
            results[prot]=0
        # loop over all the experiments to count the peptides 
        for prot in unique_proteins:
            for exp_name in self.get_experimental_names():
                if self[exp_name].is_a_parent_protein(prot):
                    results[prot]+=1
        # return the results after filling it 
        return results  
    
    def compute_protein_coverage_over_the_set(self)->Dict[str, np.ndarray]:
        """
        :return: The mapped representation for each protein accross the entire set
        :rtype: Dict[str, np.ndarray]
        """
        results:Dict[str, np.ndarray]=dict()
        unique_proteins=self.get_unique_proteins()
        # get a consent representation 
        for prot in unique_proteins:
            for exp_name in self.get_experimental_names():
                if self[exp_name].is_a_parent_protein(prot):
                    if prot in results.keys(): 
                        results[prot]+=self[exp_name].get_mapped_protein(prot)
                    else:
                        results[prot]=self[exp_name].get_mapped_protein(prot)
        # return the results after filling the array 
        return results


    def compute_correlation_in_experssion(self)->pd.DataFrame:
        """computes the correlation in parent protein gene-expression across all the experiments
        in the set. See the function **compute_binary_correlation** in the analysis module 
        for information about the computational logic.

        :return: returns a 2D matrix containing the coorelation in gene expression between each pair of experiments inside the current instance collection of experiments.  
        :rtype: pd.DataFrame
        """
        # get the experimental index
        exps_ids: List[str] = self.get_experimental_names()  
        # allocate an array to hold the results
        res_array: np.ndarray = np.zeros((len(exps_ids),len(exps_ids)))
        # fill the array with the expression correlation 
        for row_idx in range(len(exps_ids)): 
            for col_idx in range(len(exps_ids)):
                res_array[row_idx,col_idx] = compute_expression_correlation(
                    self.get_experiment(exps_ids[row_idx]),
                    self.get_experiment(exps_ids[col_idx])
                )
        # create a dataframe 
        res: pd.DataFrame =pd.DataFrame(res_array)
        # add the col and row names 
        res.columns=exps_ids
        res.index=exps_ids
        # return the results 
        return res 

    def compute_change_in_protein_representation(self)->np.ndarray:
        """Computes the change in protein representation among the proteins that are presented/ detected in all of the \
        instance's experiments.
        
        :returns: a 3D tensor, T, with shape of (num-experiments, num-experiments, num-proteins), \
        where T[i,j,k] is a the difference between experiment i & j with respect to the kth protein \
        :rtype: np.ndarray
        """
        # get the number of experiments and proteins 
        present_in_all: List[str] = self.get_proteins_present_in_all()
        num_exps: int = self.get_num_experiments_in_the_set()
        # allocate an array to hold the results 
        results_array: np.ndarray = np.zeros(shape=(num_exps, num_exps,len(present_in_all)))
        ## fill the array 
        # create some counters 
        col_counter: int = 0
        row_counter: int= 0
        for prod_idx in range(len(present_in_all)): 
            for exp_col in self.get_experiments().keys():
                for exp_row in self.get_experiments().keys():
                    results_array[row_counter,col_counter,prod_idx]=compute_change_in_protein_representation(
                            self.get_experiment(exp_row).get_mapped_protein(present_in_all[prod_idx]), 
                            self.get_experiment(exp_col).get_mapped_protein(present_in_all[prod_idx])
                        )
                    # increase the row counters 
                    row_counter+=1
                row_counter=0
                # increase the columns counter
                col_counter+=1 
            col_counter=0
        # return the results
        return results_array

    def compute_average_distance_between_exps(self)->pd.DataFrame: 
        """computes the average distance between experiments by taking the average over the z-axis
        of the 3D tensor computed by the function compute_change_in_protein_representation.
        
        :return:  A 2D tensor with shape of (num-experiments, num-experiments)
        :rtype: pd.DataFrame
        """
        diff_protein_overlap: np.ndarray = self.compute_change_in_protein_representation()
        # average overlap is: 
        average_scores: np.ndarray = np.sum(diff_protein_overlap,axis=-1)
        # construct a dataframe of the results 
        results_df: pd.DataFrame = pd.DataFrame(average_scores)
        # add the colnames 
        results_df.columns=list(self.get_experiments().keys())
        results_df.index=list(self.get_experiments().keys())
        # return the results 
        return results_df
    
    def compare_peptide_counts(self)->pd.DataFrame:
        """
        :return: A table that contain the total number of peptides and per-organism peptide counts \
        among all experiments in the set
        :rtype: pd.DataFrame
        """
        # allocate a list to hold the results 
        total_count: List[str]= []
        experiment_name: List[str]= []
        # fill the lists 
        for name in self.get_experimental_names():
            experiment_name.append(name)
            total_count.append(len(self[name].get_peptides()))
        # create a dataframe to store the results 
        res: pd.DataFrame = pd.DataFrame({
            'Organisms':['Total']*len(total_count), 
            'Counts':total_count,
            'Names':experiment_name
        })
        # get the count per organism in each experiment 
        for exp_name in self.get_experimental_names():
            # get the temp results 
            temp_res: pd.DataFrame = self[exp_name].get_peptides_per_organism()
            # add the experiment name as an extra column 
            temp_res['Names'] = [exp_name] * temp_res.shape[0]
            # concatenate the results
            res= pd.concat([res,temp_res], axis=0)
        # return the results 
        return res
    
    def compute_peptide_length_table(self)->pd.DataFrame:
        """
        :return: A table that contain the length of each peptide in the experiment
        :rtype: pd.DataFrame
        """
        # allocate a data frame to hold the results 
        res: pd.DataFrame = pd.DataFrame(columns=['Peptide_length','Names'])
        # loop over all the experiment
        for exp_name in self.get_experimental_names():
            temp_res: pd.DataFrame = pd.DataFrame({
                'Peptide_length': self[exp_name].get_peptides_length(),
                'Names': [exp_name]*len(self[exp_name].get_peptides_length())
            }) 
            res= pd.concat([res, temp_res],axis=0)
        # return the results 
        return res

    
   

















#!/usr/bin/env python 
"""This module provides an abstraction for an IP experiment. 
"""
# load the modules 
from __future__ import annotations
import pandas as pd
import numpy as np 
from IPTK.Classes.Peptide import Peptide
from IPTK.Classes.Protein import Protein
from IPTK.Classes.Proband import Proband
from IPTK.Classes.HLASet import HLASet
from IPTK.Classes.Tissue import Tissue
from IPTK.Classes.Database import SeqDB, OrganismDB
from IPTK.Utils.Mapping import map_from_uniprot_gene
from IPTK.Utils.Types import Sequences, MappedProtein, MappedProteins,ProteinSource
from typing import List, Dict, Set, Tuple
# define the analysis types 
Peptides=List[Peptide]
Proteins=List[Protein]
# define the experiment class 
class Experiment: 
	"""A representation of an immunopeptidomic experiment. 
	"""
	def __init__(self, proband:Proband, hla_set:HLASet, tissue:Tissue,  database:SeqDB,
				ident_table:pd.DataFrame)->Experiment: 
		"""Constructs an Experiment instance.

		:param proband: A proband instance containing the proband, name& other meta-data. 
		:type proband: Proband
		:param hla_set: an HLASet instance containing the set of alleles from which the peptide was eluted.
		:type hla_set: HLASet
		:param tissue: an instance of type tissue containing expression values and protein location from the corresponding tissue.
		:type tissue: Tissue
		:param database: a sequence database to exact the sequences of the identified proteins.
		:type database: SeqDB
		:param ident_table: The identification table which contain the peptides inferred from analyzing raw mass spec files.  
		:type ident_table: pd.DataFrame
		:raises ValueError: incase the provided input does not match its proposed type. 
		:return: an experiment instance 
		:rtype: Experiment
		"""
		# check the type of the inputs 
		if not isinstance(proband, Proband):
			raise ValueError(f'proband must be a Proband instance, however, your input is of type: {type(proband)}')
		
		if not isinstance(hla_set,HLASet):
			raise ValueError(f'hla_set must be an HLASet instance, however, your input is of type: {type(hla_set)}')

		if not isinstance(database, SeqDB):
			raise ValueError(f'database must be a DataBase instance, however, your input is of type: {type(database)}')

		if not isinstance(tissue, Tissue):
			raise ValueError(f'tissue must be a Tissue instance, however, your input is of type: {type(SeqDB)}')

		if not isinstance(ident_table, pd.DataFrame):
			raise ValueError(f'ident_table must be a pd.DataFrame instance, however, your input is of type: {type(ident_table)}')	
		
		if ident_table.shape[0]==0:
			raise ValueError('The identification tables can not be an empty table')

		if ident_table.shape[1] != 4:
			raise ValueError('The identification table must have 4 columns see IO modules for more details')

		# add the variables 
		self._proband: Proband =proband
		self._tissue: Tissue =tissue
		self._database: SeqDB =database
		self._hla_set: HLASet =hla_set
		# store the identified peptides as a dict object with peptide sequences as keys and Peptide instance as tables 
		self._peptides: Dict[str,Peptide] =dict()
		# store the identified proteins as set of protein ids 
		self._proteins: List[str]=list(set(ident_table.iloc[:,1].tolist()))
		# fill the instance dict 
		for idx in range(ident_table.shape[0]): 
			# if the peptide is not in the peptide dictionary we add to it. 
			if ident_table.iloc[idx,0] not in self._peptides.keys():
				self._peptides[ident_table.iloc[idx,0]]= Peptide(ident_table.iloc[idx,0])
			# link the parent protein to its child peptide. 
			self._peptides[ident_table.iloc[idx,0]].add_parent_protein(
				Protein(ident_table.iloc[idx,1], self._database.get_seq(ident_table.iloc[idx,1])), 
				ident_table.iloc[idx,2],ident_table.iloc[idx,3])
		# the instance sequence has been filled 
	
	def get_peptides_length(self)->List[int]: 
		"""return a list containing the length of each unique peptide in the database.

		:return: peptides' lengths 
		:rtype: List[int]
		"""
		return [len(pep) for pep in self.get_peptides()]

	def annotate_proteins(self, organisms_db: OrganismDB)->None:
		"""Extract the parent organisms of each protein in the experiment from an organism database instance. 

		:param organisms_db: an OrgansimDB instance that will be used to annotate the proteins \
		identified in the experiment.
		:type organisms_db: OrganismDB
		"""
		# get the organism
		proteins: List[str]= self.get_proteins()
		# allocate a list to store the parent proteins source organisms
		orgs: Dict[str,str] = dict()
		#  fill the list of elements in the list 
		for protein in proteins:
			orgs[protein]=organisms_db.get_org(protein) # This might rise a KeyError incase the protein is not in the database
		# add the organism info 
		self.add_org_info(orgs)
		return 	

	def get_orgs(self)->List[str]:
		"""return a list containing the UNIQUE organisms identified in the current experiment 

		:return: list of all UNIQUE organisms inferred from the inferred proteins.   
		:rtype: List[str]
		"""
		# allocate a list to hold the results
		unique_results: List[str]=[]
		# get the organisms
		for peptide in self.get_peptides():
			unique_results.extend(self._peptides[peptide].get_parents_org())
		# create a set the contain the unique proteins
		unique_results = list(set(unique_results))
		# return the results 
		return unique_results

	def get_peptides_per_organism(self)->pd.DataFrame:
		"""return a pandas dataframe that contain the count of peptides belonging to each organism in
		the database

		:return: a table with two columns, namely, Organisms and Counts
		:rtype: pd.DataFrame
		"""
		# allocate a list to hold the organims per proteins
		peptides_per_organims : Dict[str, int] = dict()
		# get the number of organims 
		organisms: List[str] = self.get_orgs()
		# initialize the counter with zeros 
		for org in organisms:
			 peptides_per_organims[org] = 0
		# obtain the data from the peptides 
		for pep in self.get_peptides(): 
			for org in self._peptides[pep].get_parents_org():
				peptides_per_organims[org]+=1 # increment the counter 
		# make the data compatible with data frames 
		for org in peptides_per_organims.keys(): 
			peptides_per_organims[org] = [peptides_per_organims[org]]
		# create a dataframe
		res: pd.DataFrame = pd.DataFrame(peptides_per_organims).T
		# add the index as an extra-columns 
		res['Organisms'] = res.index.tolist()
		res.columns=['Counts','Organisms']
		# reformat the dataframe 
		res.reset_index(drop=True,inplace=True)
		# swap the columns
		res=res.reindex(columns=['Organisms','Counts'])
		# sort the results 
		res=res.sort_values(by='Counts',ascending=False)
		return res
	
	def drop_peptide_belong_to_org(self, org:str)->None: 
		"""Drop the all the peptides that belong to a user provided organism. \
		Note that, this function will IRREVERSIBLY remove the peptide from the experimental object. 

		:param org: the organims name
		:type org: str
		"""
		# remove the peptides belonging to the provided organisms
		for pep in self.get_peptides():
			if org in self._peptides[pep].get_parents_org():
				self._peptides.pop(pep) # remove the peptide from the database 
		# update the list of proteins in the database by dropping proteins with no associated peptide 
		new_proteins: List[str] = []
		for pep in self.get_peptides():
			new_proteins.extend(self._peptides[pep].get_parent_proteins())
		# update the list of parent proteins
		self._proteins=list(set(new_proteins))
		return 
	
	def get_experiment_reference_tissue_expression(self) ->pd.DataFrame:
		"""return the reference gene expression for the current tissue 
	
		:return: A table that contain the expression value for ALL the genes in the instance Tissue 
		:rtype: pd.DataFrame
		"""
		return self._tissue.get_expression_profile().get_table()

	def add_org_info(self, prot2org: ProteinSource)->None:
		"""annotated the inferred proteins with their source organism

		:param prot2org: a dict that contain the protein id as keys and its source organism as values \
		and add this info to each protein inferred in the current experiment.
		:type prot2org: ProteinSource
		:raises RuntimeWarning: If the provided dictionary does cover all proteins in the experimental object.
		"""
		if len(prot2org) < len(self._proteins):
			raise RuntimeWarning(f'The provided dictionary does cover all proteins in the experimental object')
		# add the info to each parent protein
		for prot in prot2org.keys(): 
			for pep in self._peptides.keys():
				if self._peptides[pep].is_child_of(prot):
					self._peptides[pep].add_org_2_parent(prot, prot2org[prot])
		return
	
	def get_flanked_peptides(self,flank_length: int) -> Sequences:
		"""returns a list of sequences containing the peptides identified in the experiment padded with
		the flanking regions from all the parents of each peptide.

		:param flank_length: the length of the flanking region 
		:type flank_length: int
		:return: a list of the peptides + the flanking region. 
		:rtype: Sequences
		"""
		results: List[str]=[]
		for pep_idx in self._peptides.keys(): 
			results.extend(self._peptides[pep_idx].get_flanked_peptide(flank_length))
		return results
	
	def get_negative_example(self, fold: int= 2)->Sequences: 
		"""generate negative examples, i.e., non-bounding peptides from the proteins identified in the current experiment.  

		:param fold: the number of negative example to generate relative to the number of unique identified peptides, defaults to 2
		:type fold: int, optional
		:return: list of non-presented peptides from all inferred proteins. 
		:rtype: Sequences
		"""
		results: List[str]=[]
		for _ in range(fold): 
			for pep_idx in self._peptides.keys(): 
				results.extend(self._peptides[pep_idx].get_non_presented_peptides(len(pep_idx)))
		return results
 
	def get_binarized_results(self)->MappedProtein: 
		"""Return a list of NumPy arrays where each array represents a child peptide, parent protein mapped pair.
		Please note that, The function treat each peptide-protein pair individually, that is if two peptides originating from the same protein, 
		it treat them independently and the same protein will be represented twice with the two different peptides. Incase an integrative mapping is needed,
		the function @get_integrated_binarized_results@ shall be used.
		
		:return: a list of NumPy arrays containing the mapping between each peptide protein pair. 
		:rtype: MappedProtein
		"""
		results=[]
		for pep_idx in self._peptides.keys(): 
			results.extend(self._peptides[pep_idx].map_to_parent_protein())
		return results 
	
	def get_peptide(self, pep_seq: str) -> Peptide:
		"""return a peptide instance corresponding to the user provided peptide sequence.
	
		:param pep_seq: the peptide sequence 
		:type pep_seq: str
		:raises KeyError: if the peptide sequence has not been inferred from the current database. 
		:return: the peptide instance with the coresponding sequence
		:rtype: Peptide
		"""
		if pep_seq not in self._peptides.keys(): 
			raise KeyError(f"your provided petide sequence: {pep_seq} is not defined in the current experiment.") 
		return self._peptides[pep_seq]
	
	def get_mapped_protein(self, pro_id: str)->np.ndarray: 
		"""return an NumPy array of shape 1 x protein length where each number in the array represents 
		the total number of peptides identified in the experiment that have originated from the said position 
		in the protein.
		
		:param pro_id: the protein id
		:type pro_id: str
		:raises KeyError: if the provided protein id was inferred from the current experiment
		:return: a NumPy array that contain the coverage of the protein.
		:rtype: np.ndarray
		"""
		# check if the protein have been identified in the experiment 
		if pro_id not in self._proteins:
			raise KeyError(f"The provided id: {pro_id} is not in the set of identified proteins")
		# store the index of the start and end position of the child peptides   
		start_idxs, end_idxs= [], []
		# obtain all the peptides and a protein instance          
		for pep in self._peptides.keys():
			if self._peptides[pep].is_child_of(pro_id):
				protein=self._peptides[pep].get_parent(pro_id)
				start,end=self._peptides[pep].get_pos_in_parent(pro_id)
				start_idxs.append(start)
				end_idxs.append(end)
		# return the results 
		return protein.get_peptides_map(start_idxs,end_idxs)
		
	def get_expression_of_parent_proteins(self, non_mapped_dval: float = -1)->pd.DataFrame:
		"""return a table containing the expression of the proteins inferred in the current experiment from the current tissue.
		This method need internet connection as it need to access uniprot mapping API to map uniprot IDs to gene IDs.  

		:param non_mapped_dval: A default value to be added incase the parent protein is not define in the expression database, defaults to -1
		:type non_mapped_dval: float, optional
		:return: a table that contain the expression of the protein inferred in the database 
		:rtype: pd.DataFrame
		"""
		proteins: List[str] = list(self.get_proteins())
		map2Ensemble: pd.DataFrame = map_from_uniprot_gene(proteins)
		# allocate a list to hold the expression values 
		expression: List[float] = []
		for prot in proteins:
			# get  all transcripts that map to the protein -> port
			temp_df: pd.DataFrame = map2Ensemble.loc[map2Ensemble.iloc[:,0]==prot]
			# if one-to-one mapping is returned 
			if temp_df.shape[0] ==1:
				# we try to extract the expression value of the protein
				try:
					expression.append(self._tissue.get_expression_profile().get_gene_id_expression(temp_df.iloc[0,1]))
				except KeyError:
					expression.append(non_mapped_dval)
			else: # we have more than one mapping 
				temp_ens_ids: List[str] =  temp_df.iloc[:,1].tolist()
				temp_res_raw: List[int] = []
				for ens_id in temp_ens_ids:
					try: 
						temp_res_raw.append(self._tissue.get_expression_profile().get_gene_id_expression(ens_id))
					except KeyError:
						temp_res_raw.append(non_mapped_dval)
				# filter out default value 
				temp_res_pross: List[int] = [elem for elem in temp_res_raw if elem != non_mapped_dval]
				# if the list is empty, all the transcript can not be mapped 
				if len(temp_res_pross)==0:
					expression.append(non_mapped_dval)
				else: 
					expression.append(np.mean(temp_res_pross))	 
		# construct the dataframe 
		results: pd.DataFrame= pd.DataFrame({'Proteins':proteins, 'Expression':expression})
		return results
	
	def get_main_sub_cellular_location_of_parent_proteins(self, not_mapped_val: str = 'UNK')->pd.DataFrame:
		"""retrun the main cellular location for the identified proteins.
		This method need internet connection as it need to access uniprot mapping API to map uniprot IDs to gene IDs.  


		:param not_mapped_val: The default value to return incase the location of a protein can not be extracted, defaults to 'UNK'
		:type not_mapped_val: str, optional
		:return: A table that contain the main cellular compartment for each protein in the current instance.
		:rtype: pd.DataFrame
		"""
		proteins: List[str] = list(self.get_proteins())
		map2Ensemble: pd.DataFrame = map_from_uniprot_gene(proteins)
		# allocate a list to hold the main location
		main_locations: List[str] = []
		for prot in proteins:
			# we get a pandas dataframe that contain all the ensemble ids belonging to this protein.  
			temp_df: pd.DataFrame = map2Ensemble.loc[map2Ensemble.iloc[:,0]==prot]
			if temp_df.shape[0]==1: 
				try:
					main_locations.append(';'.join(self._tissue.get_subCellular_locations().get_main_location(temp_df.iloc[0,1])))
				except KeyError: 
					main_locations.append(not_mapped_val)
			else: 
				temp_ens_ids: List[str] = temp_df.iloc[:,1].tolist()
				temp_res_raw: List[str] = []
				for ens_id in temp_ens_ids:
					try: 
						temp_res_raw.append(';'.join(self._tissue.get_subCellular_locations().get_main_location(ens_id)))
					except KeyError: 
						temp_res_raw.append(not_mapped_val)
				# filter out default value 
				temp_res_pross: List[int] = [elem for elem in temp_res_raw if elem != not_mapped_val]
				# if the list is empty, all the proteins can not be mapped 
				if len(temp_res_pross)==0:
					main_locations.append(not_mapped_val)
				else: 
					# get a set of the unique location from different mapping 
					temp_unique_poss: Set[str] = set()
					for elem in temp_res_pross: 
						for loc in elem.split(';'):
							temp_unique_poss.add(loc)
					# append the results into one string and add it to the database elements 		
					main_locations.append(';'.join(temp_unique_poss))
		# construct the dataframe 
		results: pd.DataFrame= pd.DataFrame({'Proteins':proteins, 'Main_locations':main_locations})
		return results

	
	def get_go_location_id_parent_proteins(self, not_mapped_val: str = 'UNK')->pd.DataFrame:
		"""retrun the gene ontology,GO, location terms for all the identified proteins. 

		:param not_mapped_val: The default value to return incase the GO term of the protein can not be extracted, defaults to 'UNK'
		:type not_mapped_val: str, optional
		:return: A table that contain the GO-location term for each protein in the current instance.
		:rtype: pd.DataFrame
		"""
		proteins: List[str] = list(self.get_proteins())
		map2Ensemble: pd.DataFrame = map_from_uniprot_gene(proteins)
		#allocate a list to hold the go terms
		go_terms: List[str] = []
		for prot in proteins:
			# we get a pandas dataframe that contain all the ensemble ids belonging to this protein.  
			temp_df: pd.DataFrame = map2Ensemble.loc[map2Ensemble.iloc[:,0]==prot]
			if temp_df.shape[0]==1: 
				try:
					go_terms.append(';'.join(self._tissue.get_subCellular_locations().get_go_names(temp_df.iloc[0,1])))
				except KeyError: 
					go_terms.append(not_mapped_val)
			else: 
				temp_ens_ids: List[str] = temp_df.iloc[:,1].tolist()
				temp_res_raw: List[str] = []
				for ens_id in temp_ens_ids:
					try: 
						temp_res_raw.append(';'.join(self._tissue.get_subCellular_locations().get_go_names(ens_id)))
					except KeyError: 
						temp_res_raw.append(not_mapped_val)
				# filter out default value 
				temp_res_pross: List[int] = [elem for elem in temp_res_raw if elem != not_mapped_val]
				# if the list is empty, all the proteins can not be mapped 
				if len(temp_res_pross)==0:
					go_terms.append(not_mapped_val)
				else: 
					# get a set of the unique location from different mapping 
					temp_unique_poss: Set[str] = set()
					for elem in temp_res_pross: 
						for loc in elem.split(';'):
							temp_unique_poss.add(loc)
					# append the results into one string and add it to the database elements 		
					go_terms.append(';'.join(temp_unique_poss))
		# construct the dataframe 
		results: pd.DataFrame= pd.DataFrame({'Proteins':proteins, 'GO_Terms':go_terms})
		return results

	def get_num_peptide_expression_table(self)->pd.DataFrame:
		"""Get a table that contain the id of all parent proteins, number of peptide per-proteins and the expression value 
		   of these parent transcripts. Please note, this method need internet connection as it need to access uniprot mapping API to map uniprot IDs to gene IDs.  

		:return: the number of peptides per protein table
		:rtype: pd.DataFrame
		"""
		# get the number of tables per peptides 
		num_peptides_per_protein: pd.DataFrame = self.get_peptides_per_protein()
		expression_level: pd.DataFrame = self.get_expression_of_parent_proteins()
		# merge the tables 
		results: pd.DataFrame =pd.merge(num_peptides_per_protein,expression_level)
		# return the results 
		return results

	def get_number_of_proteins_per_compartment(self) -> pd.DataFrame: 
		"""returns the number of proteins from each compartment 
		
		:return: A table that has two columns, namely, Compartment and Counts. 
		:rtype: pd.DataFrame
		"""
		# get the main locations: 
		parent_protein_locations: pd.DataFrame = self.get_main_sub_cellular_location_of_parent_proteins()
		# obtain the locations as a list 
		locations: List[str] = []
		for loc in parent_protein_locations.iloc[:,1].tolist(): 
			locations.extend(loc.split(';'))
		# construct a dictionary to hold the counts 
		unique_compartments: Set[str] = set(locations)
		compartment_counts: Dict[str,int] = dict()
		# initialize the counts 
		for comp in unique_compartments: 
			compartment_counts[comp]=0
		# update the counter 
		for loc in locations: 
			compartment_counts[loc]+=1
		# prepare the dict to be compatible with dataframe 
		for key in  compartment_counts.keys():
			compartment_counts[key]= [compartment_counts[key]]
		# construct a data frame from the results 
		res: pd.DataFrame = pd.DataFrame(compartment_counts).T
		# add the index as an extra-columns 
		res['Compartment'] = res.index.tolist()
		res.columns=['Counts','Compartment']
		# reformat the dataframe 
		res.reset_index(drop=True,inplace=True)
		# swap the columns
		res=res.reindex(columns=['Compartment','Counts'])
		# sort the results 
		res=res.sort_values(by='Counts',ascending=False)
		return res

	def get_number_of_proteins_per_go_term(self) -> pd.DataFrame: 
		"""returns the number of proteins from each GO-Term 
		
		:return: A table that has two columns, namely, GO-Terms and Counts. 
		:rtype: pd.DataFrame
		"""
		# get the go-terms
		parent_protein_go_term : pd.DataFrame = self.get_go_location_id_parent_proteins()
		# obtain the locations as a list 
		go_terms: List[str] = []
		for terms in parent_protein_go_term.iloc[:,1].tolist(): 
			go_terms.extend(terms.split(';'))
		# construct a dictionary to hold the counts 
		unique_terms: Set[str] = set(go_terms)
		terms_counts: Dict[str,int] = dict()
		# initialize the counts 
		for term in unique_terms: 
			terms_counts[term]=0
		# update the counter 
		for term in go_terms: 
			terms_counts[term]+=1
		# prepare the dict to be compatible with dataframe 
		for key in  terms_counts.keys():
			terms_counts[key]= [terms_counts[key]]
		# construct a data frame from the results 
		res: pd.DataFrame= pd.DataFrame(terms_counts).T
		# add the terms as a column
		res['Terms']= res.index.tolist()
		res.columns=['Counts','GO-Terms']
		# reformat the dataframe 
		res.reset_index(drop=True,inplace=True)
		# swap the columns
		res=res.reindex(columns=['GO-Terms','Counts'])
		# sort the results 
		res=res.sort_values(by='Counts',ascending=False)
		return res

	def get_num_peptide_per_location(self)->pd.DataFrame:
		"""retruns the number of peptides obtained from proteins localized to different sub-cellular compartments  
		
		:return: A table that has two columns, namely, Compartment and Counts.
		:rtype: pd.DataFrame
		"""
		# get unique compartments 
		unique_locations: List[str]= self.get_number_of_proteins_per_compartment().iloc[:,0].tolist()
		# initialize the counter to zeros
		pep_per_loc: Dict[str,int]=dict()
		for loc in unique_locations:
			pep_per_loc[loc]=0
		# get the location of each parent protein 
		parent_protein_locs: pd.DataFrame = self.get_main_sub_cellular_location_of_parent_proteins()
		peptide_count_parents: pd.DataFrame = self.get_peptides_per_protein()
		# loop over the parent proteins and update the countert 
		for idx in range(parent_protein_locs.shape[0]):
			# get the number of peptides belonging to this protein  
			num_peptides: int = peptide_count_parents.loc[peptide_count_parents.iloc[:,0]==parent_protein_locs.iloc[idx,0]]['Number_of_Peptides'].tolist()[0]
			# get the locations 
			locations: List[str] = parent_protein_locs.iloc[idx,1].split(';')
			# add the locations to the list 
			for loc in locations:
				pep_per_loc[loc]+=num_peptides
		# make the dict compatible with dataframes 
		for key in pep_per_loc.keys():
			pep_per_loc[key]=[pep_per_loc[key]]
		# construct a data frame from the results 
		res: pd.DataFrame = pd.DataFrame(pep_per_loc).T
		# add the index as a columns 
		res['Compartment']=res.index.tolist()
		res.columns=['Counts','Compartment']
		# reformat the dataframe 
		res.reset_index(drop=True,inplace=True)
		# swap the columns
		res=res.reindex(columns=['Compartment','Counts'])
		# sort the results 
		res=res.sort_values(by='Counts',ascending=False)
		# return the results 
		return res 

	def get_num_peptide_per_go_term(self)->pd.DataFrame:
		"""retruns the number of peptides per each GO-Term 
		:return: A table that has two columns, namely, GO-Terms and Counts. 
		:rtype: pd.DataFrame
		"""
		# get GO terms 
		unique_go_terms: List[str]= self.get_number_of_proteins_per_go_term().iloc[:,0].tolist()
		# initialize the counter to zeros
		pep_per_term: Dict[str,int]=dict()  
		for loc in unique_go_terms:
			pep_per_term[loc]=0
		# get the Go-Terms of each parent protein 
		parent_protein_go_terms: pd.DataFrame = self.get_go_location_id_parent_proteins()
		peptide_count_parents: pd.DataFrame = self.get_peptides_per_protein()
		# loop over the parent proteins and update the countert 
		for idx in range(parent_protein_go_terms.shape[0]):
			# get the number of peptides belonging to this protein  
			num_peptides: int = peptide_count_parents.loc[peptide_count_parents.iloc[:,0]==parent_protein_go_terms.iloc[idx,0]]['Number_of_Peptides'].tolist()[0]
			# get the locations 
			go_terms: List[str] = parent_protein_go_terms.iloc[idx,1].split(';')
			# add the locations to the list 
			for term in go_terms:
				pep_per_term[term]+=num_peptides
		# make the dict compatible with dataframes 
		for key in pep_per_term.keys():
			pep_per_term[key]=[pep_per_term[key]]
		# construct a data frame from the results 
		res: pd.DataFrame = pd.DataFrame(pep_per_term).T
		# add the index as a columns 
		res['Compartment']=res.index.tolist()
		res.columns=['Counts','GO-Terms']
		# reformat the dataframe 
		res.reset_index(drop=True,inplace=True)
		# swap the columns
		res=res.reindex(columns=['GO-Terms','Counts'])
		# sort the results 
		res=res.sort_values(by='Counts',ascending=False)
		# return the results 
		return res 

	def get_mapped_proteins(self)->MappedProteins: 
		"""returns a dictionary of all the proteins identified in the current experiment with all inferred
		peptides mapped to them. 

		:return: a dictionary that contain the mapped proteins for all the proteins in the current instance. 
		:rtype: MappedProteins
		"""
		results=dict()
		for prot in self._proteins:
			results[prot]=self.get_mapped_protein(prot)
		return results 
	
	def get_mono_parent_peptides(self)->Peptides:
		"""
		returns a list of peptides that have only one parent protein
		
		:return: list of peptide instance 
		:rtype: Peptides
		"""
		results: List[Peptide]=[]
		for pep in self._peptides.keys():
			if self._peptides[pep].get_number_of_parents()==1:
				results.append(self._peptides[pep])
		return results
	
	def get_poly_parental_peptides(self)->Peptides:	
		"""returns a list of peptides that have more than one parent protein
		:return: [list of peptide instance 
		:rtype: Peptides
		"""
		results: List[Peptide]=[]
		for pep in self._peptides.keys():
			if self._peptides[pep].get_number_of_parents()>1:
				results.append(self._peptides[pep])
		return results
	
	def get_peptide_number_parent(self, ascending: bool = False)-> pd.DataFrame:
		"""returns a pandas dataframe with the peptide sequence in the first columns and the 
		number of parent proteins in the second column. 
		
		:param ascending: ascending sort the peptide by their number of parent proteins, defaults to False
		:type ascending: bool, optional
		:return: the number of parents for each peptide 
		:rtype: pd.DataFrame
		"""
		peptides=list(self._peptides.keys())
		num_parents=[]
		for second in peptides: 
			num_parents.append(self._peptides[second].get_number_of_parents())
		# combine the results into a dataframe 
		res=pd.DataFrame({'Peptides':peptides,'Number_of_parents':num_parents})
		# sort the results 
		res.sort_values(axis=0, by='Number_of_parents', ascending=ascending, inplace=True)
		# return the results 
		return res 
	
	def get_number_of_children(self,pro_id: str)->int:
		"""returns the number of children, i.e. number of peptides belonging to a parent protein

		:param pro_id: the id of the parent protein 
		:type pro_id: str
		:return: the number of peptides
		:rtype: int
		"""
		children=0
		for pep in self._peptides.keys(): 
			if self._peptides[pep].is_child_of(pro_id):
				children+=1
		return children
	
	def get_peptides_per_protein(self, ascending: bool = False)->pd.DataFrame:
		"""returns a pandas dataframe that contain the number of peptides belonging to each protein 
		inferred in the experiment
		
		:param ascending: ascending sort the proteins by their number of parents each child peptide has, defaults to False
		:type ascending: bool, optional
		:return: A table with the following columns, Proteins and Number_of_Peptides
		:rtype: pd.DataFrame
		"""
		proteins=list(self._proteins)
		number_peptides=[]
		for prot in proteins: 
			number_peptides.append(self.get_number_of_children(prot))
		# construct a pandas dataframe with the proteins and number of peptides 
		res=pd.DataFrame({'Proteins':proteins,'Number_of_Peptides':number_peptides})
		# sort the resulting table 
		res.sort_values(axis=0, by='Number_of_Peptides',ascending=ascending, inplace=True)
		# return the results 
		return res 

	def get_n_terminal_flanked_seqs(self, flank_length: int)->Peptides:
		"""returns the n-terminal flanking sequences 

		:param flank_length: the length of the flanking region upstream of the N-terminal of the peptide. 
		:type flank_length: int
		:return: a list of sequences containing the N-terminal flanking sequence for each peptide in the instance.  
		:rtype: Peptides
		"""
		results=[]
		for pep_id in self._peptides.keys(): 
			results.extend(self._peptides[pep_id].get_n_terminal_flank_seq(flank_length))
		return results 

	def get_c_terminal_flanked_seqs(self, flank_length: int)-> Peptides:
		"""returns the c-terminal flanking sequences

		:param flank_length: the length of the peptide downstream of the C-terminal of the peptide  
		:type flank_length: int
		:return: a list of sequences containing the N-terminal flanking sequence for each peptide in the instance.  
		:rtype: Peptides
		"""
		results=[]
		for pep_id in self._peptides.keys(): 
			results.extend(self._peptides[pep_id].get_c_terminal_flank_seq(flank_length))
		return results 

	def get_tissue_name(self)->str:
		"""
		:returns: the name of the tissue 
		:rtype: str
		"""
		return self._tissue.get_name()
		
	def get_proband_name(self)->str:
		"""
		:return: the proband's name
		:rtype: str
		"""
		return self._proband.get_name()
	
	def get_hla_class(self)->int:
		"""
		:return: the HLA class
		:rtype: int
		"""
		return self._hla_set.get_class()
	
	def get_hla_allele(self)-> List[str]:
		"""
		:return: the set of HLA alleles from which the instance peptides have been eluted 
		:rtype: List[str]
		"""
		return self._hla_set.get_alleles()
		
	def has_hla_allele(self, individual: str)->bool:
		"""returns whether or not the experiment contain an eluted peptides from the provided alleles 
		
		:param individual: the name of the allele as a string
		:type individual: str
		:return: True if the allele is a member of the instance HLASet and False otherwise.  
		:rtype: bool
		"""
		return self._hla_set.has_allele(individual)
	
	def has_gene(self, locus: str)->bool:
		"""returns whether or not the experiment contains peptides eluted from an HLA-allele belonging to the provided locus or not
		
		:param locus: the locus of the allele to query the hla_set against
		:type locus: str
		:return: True if the locus has a member that is a member of the instance HLASet and False otherwise
		:rtype: bool
		"""
		return self._hla_set.has_gene(locus)
	
	def has_allele_group(self, gene_group:str)->bool:
		"""returns whether or not the experiment contains peptides eluted from an HLA-allele belonging to the provided allele group or not
		
		:param gene_group: the gene group to query the hla_set against 
		:type gene_group: str
		:return: True if the gene group has a member that is a member of the instance HLASet and False otherwise
		:rtype: bool
		"""
		return self._hla_set.has_allele_group(gene_group)
	
	def has_protein_group(self, protein_group: str)->bool:
		""" returns whether or not the experiment contains peptides eluted from an HLA-allele belonging to the provided protein group or not

		:param protein_group: The protein group to query the hla_set against 
		:type protein_group: str
		:return: True if the locus has a member that is a member of the instance HLASet and False otherwise
		:rtype: bool
		"""
		return self._hla_set.has_protein_group(protein_group)
	
	def get_peptides(self)->Peptides:
		"""
		:return: a set of all the peptides stored in the experimental object
		:rtype: Peptides
		"""
		return set(self._peptides.keys())

	def get_proteins(self)->Proteins:
		"""
		:return: a set of all the proteins in the experimental object
		:rtype: Proteins
		"""
		return set(self._proteins)
	
	def is_member(self,peptide:str)->bool:
		"""
		:param peptide: check if the peptide is a member of the instance peptides or not.
		:type peptide: str
		:return: True if the peptide has been identified in the current instance, False otherwise. 
		:rtype: bool
		"""
		if peptide in self._peptides.keys():
			return True
		return False
	
	def is_a_parent_protein(self, protein:str)->bool:
		"""
		:param protein: check if the protein is a member of the instance proteins or not.
		:type peptide: str
		:return: True if the protein has been identified in the current instance, False otherwise. 
		:rtype: bool
		"""
		if protein in self._proteins:
			return True
		return False 
	
	def get_tissue(self)->Tissue: 
		"""
		:return: the tissue of the current experiment.
		:rtype: Tissue
		"""
		return self._tissue

	def	__len__(self)->int: 
		""" A magic function for the len function, return the number of unique peptides in the experiment.

		:return: the number of unique peptides in the database
		:rtype: int
		"""
		return len(self._peptides)
	
	def __str__(self)->str:
		"""
		:return: a string representation for the tissue 
		:rtype: str
		"""
		return f"""an Experimental from proband: {self._proband.get_name()}, Tissue: {self._tissue.get_name()}
				   With an HLA Class: {self._hla_set.get_class()} With
				   {len(self)} peptide identification from {len(self._proteins)} Protein"""
	
	
	def __repr__(self)->str:
		return str(self)
	


				













		